from cli import __app_name__, empiric_cli
from typer.testing import CliRunner

runner = CliRunner()


def test_version():
    result = runner.invoke(empiric_cli.app, ["--version"])
    assert result.exit_code == 0
    assert f"{__app_name__}\n" in result.stdout



import asyncio
import logging
import time
from typing import List

import requests
from aiohttp import ClientSession
from empiric.core.entry import GenericEntry
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherInterfaceT

logger = logging.getLogger(__name__)


class TheGraphFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://api.thegraph.com/subgraphs/name/"
    SOURCE: str = "THEGRAPH"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> GenericEntry:
        if asset["source"] == "AAVE":
            url_slug = "aave/protocol-v2"
            query = f"query {{reserves(where: {{id: \"{asset['detail']['asset_address']}\"}}) {{name isActive isFrozen {asset['detail']['metric']}}}}}"
            input_decimals = 27
        else:
            raise ValueError(
                f"Unknown asset name, do not know how to query The Graph for {asset['name']}"
            )

        async with session.post(
            self.BASE_URL + url_slug, json={"query": query}
        ) as resp:
            result_json = await resp.json()
            result = result_json["data"]["reserves"][0]

            return self._construct(asset, result, input_decimals=input_decimals)

    def _fetch_pair_sync(self, asset: EmpiricSpotAsset) -> GenericEntry:
        if asset["source"] == "AAVE":
            url_slug = "aave/protocol-v2"
            query = f"query {{reserves(where: {{id: \"{asset['detail']['asset_address']}\"}}) {{name isActive isFrozen {asset['detail']['metric']}}}}}"
            input_decimals = 27
        else:
            raise ValueError(
                f"Unknown asset name, do not know how to query The Graph for {asset['name']}"
            )

        resp = requests.post(self.BASE_URL + url_slug, json={"query": query})
        result_json = resp.json()
        result = result_json["data"]["reserves"][0]
        return self._construct(asset, result, input_decimals=input_decimals)

    async def fetch(self, session: ClientSession) -> List[GenericEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "ONCHAIN":
                logger.debug(f"Skipping The Graph for non-on-chain asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[GenericEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "ONCHAIN":
                logger.debug(f"Skipping The Graph for non-on-chain asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result, input_decimals=27) -> GenericEntry:
        key = asset["key"]

        if result["name"] != asset["detail"]["asset_name"]:
            raise ValueError("invalid json")
        if not result["isActive"] is True:
            raise ValueError("invalid json")
        if not result["isFrozen"] is False:
            raise ValueError("invalid json")

        value = float(result[asset["detail"]["metric"]])
        value_int = int(value * (10 ** (asset["decimals"] - input_decimals)))
        timestamp = int(time.time())

        logger.info(f"Fetched data {value_int} for {key} from The Graph")

        return GenericEntry(
            key=key,
            value=value_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
        )



# Copyright (C) 2020 Eric Schorn, NCC Group Plc; Provided under the MIT license

# This code follows the (IETF) IRTF CFRG Verifiable Random Functions (VRFs) spec *very* closely.
# Please refer to https://tools.ietf.org/pdf/draft-irtf-cfrg-vrf-06.pdf

# https://github.com/nccgroup/draft-irtf-cfrg-vrf-06


import hashlib

# Public API


# Section 5.1. ECVRF Proving
def ecvrf_prove(sk, alpha_string):
    """
    Input:
        sk - VRF private key (32 bytes)
        alpha_string - input alpha, an octet string
    Output:
        ("VALID", pi_string) - where pi_string is the VRF proof, octet string of length ptLen+n+qLen
        (80) bytes, or ("INVALID", []) upon failure
    """
    # 1. Use sk to derive the VRF secret scalar x and the VRF public key y = x*B
    secret_scalar_x = _get_secret_scalar(sk)
    public_key_y = get_public_key(sk)

    # 2. H = ECVRF_hash_to_curve(suite_string, y, alpha_string)
    h = _ecvrf_hash_to_curve_elligator2_25519(SUITE_STRING, public_key_y, alpha_string)
    if h == "INVALID":
        return "INVALID", []

    # 3. h_string = point_to_string(H)
    h_string = _decode_point(h)
    if h_string == "INVALID":
        return "INVALID", []

    # 4. Gamma = x*H
    gamma = _scalar_multiply(p=h_string, e=secret_scalar_x)

    # 5. k = ECVRF_nonce_generation(sk, h_string)
    k = _ecvrf_nonce_generation_rfc8032(sk, h)

    # 6. c = ECVRF_hash_points(H, Gamma, k*B, k*H)
    k_b = _scalar_multiply(p=BASE, e=k)
    k_h = _scalar_multiply(p=h_string, e=k)
    c = _ecvrf_hash_points(h_string, gamma, k_b, k_h)

    # 7. s = (k + c*x) mod q
    s = (k + c * secret_scalar_x) % ORDER

    # 8. pi_string = point_to_string(Gamma) || int_to_string(c, n) || int_to_string(s, qLen)
    pi_string = (
        _encode_point(gamma)
        + int.to_bytes(c, 16, "little")
        + int.to_bytes(s, 32, "little")
    )

    if "test_dict" in globals():
        _assert_and_sample(
            [
                "secret_scalar_x",
                "public_key_y",
                "h",
                "gamma",
                "k_b",
                "k_h",
                "pi_string",
            ],
            [
                secret_scalar_x.to_bytes(32, "little"),
                public_key_y,
                h,
                _encode_point(gamma),
                _encode_point(k_b),
                _encode_point(k_h),
                pi_string,
            ],
        )

    # 9. Output pi_string
    return "VALID", pi_string


# Section 5.2. ECVRF Proof To Hash
def ecvrf_proof_to_hash(pi_string):
    """
    Input:
        pi_string - VRF proof, octet string of length ptLen+n+qLen (80) bytes
    Output:
        ("VALID", beta_string) where beta_string is the VRF hash output, octet string
        of length hLen (64) bytes, or ("INVALID", []) upon failure
    """
    # 1. D = ECVRF_decode_proof(pi_string)
    d = _ecvrf_decode_proof(pi_string)

    # 2. If D is "INVALID", output "INVALID" and stop
    if d == "INVALID":
        return "INVALID", []

    # 3. (Gamma, c, s) = D
    gamma, c, s = d

    # 4. three_string = 0x03 = int_to_string(3, 1), a single octet with value 3
    three_string = bytes([0x03])

    # 5. beta_string = Hash(suite_string || three_string || point_to_string(cofactor * Gamma))
    cofactor_gamma = _scalar_multiply(p=gamma, e=COFACTOR)  # Curve cofactor
    beta_string = _short_hash(
        SUITE_STRING + three_string + _encode_point(cofactor_gamma)
    )

    if "test_dict" in globals():
        _assert_and_sample(["beta_string"], [beta_string])

    # 6. Output beta_string
    return "VALID", beta_string


# Section 5.3. ECVRF Verifying
def ecvrf_verify(y, pi_string, alpha_string):
    """
    Input:
        y - public key, an EC point as bytes
        pi_string - VRF proof, octet string of length ptLen+n+qLen (80) bytes
        alpha_string - VRF input, octet string
    Output:
        ("VALID", beta_string), where beta_string is the VRF hash output, octet string
        of length hLen (64) bytes; or ("INVALID", []) upon failure
    """
    # Note that the API caller is expected to verify that the returned beta_string is the
    # expected one and this has a strong potential for mistakes/oversights (such as checking
    # for "VALID" but not the actual value). Production code would be better served by
    # passing in the expected beta_string and getting a simpler pass/fail in response.

    # 1. D = ECVRF_decode_proof(pi_string)
    d = _ecvrf_decode_proof(pi_string)

    # 2. If D is "INVALID", output "INVALID" and stop
    if d == "INVALID":
        return "INVALID", []

    # 3. (Gamma, c, s) = D
    gamma, c, s = d

    # 4. H = ECVRF_hash_to_curve(suite_string, y, alpha_string)
    h = _ecvrf_hash_to_curve_elligator2_25519(SUITE_STRING, y, alpha_string)
    if h == "INVALID":
        return "INVALID", []

    # 5. U = s*B - c*y
    y_point = _decode_point(y)
    h_point = _decode_point(h)
    if y_point == "INVALID" or h_point == "INVALID":
        return "INVALID", []
    s_b = _scalar_multiply(p=BASE, e=s)
    c_y = _scalar_multiply(p=y_point, e=c)
    nc_y = [PRIME - c_y[0], c_y[1]]
    u = _edwards_add(s_b, nc_y)

    # 6. V = s*H - c*Gamma
    s_h = _scalar_multiply(p=h_point, e=s)
    c_g = _scalar_multiply(p=gamma, e=c)
    nc_g = [PRIME - c_g[0], c_g[1]]
    v = _edwards_add(nc_g, s_h)

    # 7. c’ = ECVRF_hash_points(H, Gamma, U, V)
    cp = _ecvrf_hash_points(h_point, gamma, u, v)

    if "test_dict" in globals():
        _assert_and_sample(["h", "u", "v"], [h, _encode_point(u), _encode_point(v)])

    # 8. If c and c’ are equal, output ("VALID", ECVRF_proof_to_hash(pi_string)); else output "INVALID"
    if c == cp:
        return ecvrf_proof_to_hash(pi_string)  # Includes logic for VALID/INVALID
    else:
        return "INVALID", []


def get_public_key(sk):
    """Calculate and return the public_key as an encoded point string (bytes)"""
    secret_int = _get_secret_scalar(sk)
    public_point = _scalar_multiply(p=BASE, e=secret_int)
    public_string = _encode_point(public_point)
    return public_string


# Internal functions

# Section 5.4.1.2. ECVRF_hash_to_curve_elligator2_25519
def _ecvrf_hash_to_curve_elligator2_25519(suite_string, y, alpha_string):
    """
    Input:
        suite_string - a single octet specifying ECVRF ciphersuite.
        alpha_string - value to be hashed, an octet string
        y - public key, an EC point as bytes
    Output:
        H - hashed value, a finite EC point in G, or INVALID upon failure
    Fixed options:
        p = 2^255-19, the size of the finite field F, a prime, for edwards25519 and curve25519 curves
        A = 486662, Montgomery curve constant for curve25519
        cofactor = 8, the cofactor for edwards25519 and curve25519 curves
    """
    assert suite_string == SUITE_STRING
    # 1. PK_string = point_to_string(y)
    # 2. one_string = 0x01 = int_to_string(1, 1) (a single octet with value 1)
    one_string = bytes([0x01])

    # 3. hash_string = Hash(suite_string || one_string || PK_string || alpha_string )
    hash_string = _hash(suite_string + one_string + y + alpha_string)

    # 4. r_string = hash_string[0]...hash_string[31]
    r_string = bytearray(hash_string[0:32])

    # 5. oneTwentySeven_string = 0x7F = int_to_string(127, 1) (a single octet with value 127)
    one_twenty_seven_string = 0x7F  # Note: '&' wants an int, not a byte

    # 6. r_string[31] = r_string[31] & oneTwentySeven_string (this step clears the high-order bit of octet 31)
    r_string[31] = int(r_string[31] & one_twenty_seven_string)

    # 7. r = string_to_int(truncated_h_string)
    r = int.from_bytes(r_string, "little")

    # 8. u = - A / (1 + 2*(r^2) ) mod p (note: the inverse of (1+2*(r^2)) modulo p is guaranteed to exist)
    u = (PRIME - A) * _inverse(1 + 2 * (r**2)) % PRIME

    # 9. w = u * (u^2 + A*u + 1) mod p (this step evaluates the Montgomery equation for Curve25519)
    w = u * (u**2 + A * u + 1) % PRIME

    # 10. Let e equal the Legendre symbol of w and p (see note after item 16)
    e = pow(w, (PRIME - 1) // 2, PRIME)

    # 11. If e is equal to 1 then final_u = u; else final_u = (-A - u) mod p (see note after item 16)
    final_u = (e * u + (e - 1) * A * TWO_INV) % PRIME
    # Note that while the above formula makes some sense in a constant-time implementation, this
    # implementation is not intended to be constant time. Thus it could be considerably simplified.

    # 12. y_coordinate = (final_u - 1) / (final_u + 1) mod p
    y_coordinate = (final_u - 1) * _inverse(final_u + 1) % PRIME

    # 13. y_string = int_to_string (y_coordinate, 32)
    y_string = int.to_bytes(y_coordinate, 32, "little")

    # 14. H_prelim = string_to_point(h_string)
    h_prelim = _decode_point(y_string)
    if h_prelim == "INVALID":
        return "INVALID"

    # 15. Set H = cofactor * H_prelim
    h = _scalar_multiply(p=h_prelim, e=COFACTOR)  # Curve cofactor

    # 16. Output H
    h_point = _encode_point(h)

    if "test_dict" in globals():
        _assert_and_sample(
            ["r", "w", "e"],
            [r_string, int.to_bytes(w, 32, "little"), int.to_bytes(e, 32, "little")],
        )

    return h_point


# 5.4.2.2. ECVRF Nonce Generation From RFC 8032
def _ecvrf_nonce_generation_rfc8032(sk, h_string):
    """
    Input:
        sk - an ECVRF secret key as bytes
        h_string - an octet string
    Output:
        k - an integer between 0 and q-1
    """
    # 1. hashed_sk_string = Hash (sk)
    hashed_sk_string = _hash(sk)

    # 2. truncated_hashed_sk_string = hashed_sk_string[32]...hashed_sk_string[63]
    truncated_hashed_sk_string = hashed_sk_string[32:]

    # 3. k_string = Hash(truncated_hashed_sk_string || h_string)
    k_string = _hash(truncated_hashed_sk_string + h_string)

    # 4. k = string_to_int(k_string) mod q
    k = int.from_bytes(k_string, "little") % ORDER

    if "test_dict" in globals():
        _assert_and_sample(["k"], [k_string])

    return k


# Section 5.4.3. ECVRF Hash Points
def _ecvrf_hash_points(p1, p2, p3, p4):
    """
    Input:
        P1...PM - EC points in G
    Output:
        c - hash value, integer between 0 and 2^(8n)-1
    """
    # 1. two_string = 0x02 = int_to_string(2, 1), a single octet with value 2
    two_string = bytes([0x02])

    # 2. Initialize str = suite_string || two_string
    string = SUITE_STRING + two_string

    # 3. for PJ in [P1, P2, ... PM]:
    #        str = str || point_to_string(PJ)
    string = (
        string
        + _encode_point(p1)
        + _encode_point(p2)
        + _encode_point(p3)
        + _encode_point(p4)
    )

    # 4. c_string = Hash(str)
    c_string = _hash(string)

    # 5. truncated_c_string = c_string[0]...c_string[n-1]
    truncated_c_string = c_string[0:16]

    # 6. c = string_to_int(truncated_c_string)
    c = int.from_bytes(truncated_c_string, "little")

    # 7. Output c
    return c


# Section 5.4.4. ECVRF Decode Proof
def _ecvrf_decode_proof(pi_string):
    """
    Input:
        pi_string - VRF proof, octet string (ptLen+n+qLen octets)
    Output:
        "INVALID", or Gamma - EC point
        c - integer between 0 and 2^(8n)-1
        s - integer between 0 and 2^(8qLen)-1
    """
    if len(pi_string) != 80:  # ptLen+n+qLen octets = 32+16+32 = 80
        return "INVALID"

    # 1. let gamma_string = pi_string[0]...p_string[ptLen-1]
    gamma_string = pi_string[0:32]

    # 2. let c_string = pi_string[ptLen]...pi_string[ptLen+n-1]
    c_string = pi_string[32:48]

    # 3. let s_string =pi_string[ptLen+n]...pi_string[ptLen+n+qLen-1]
    s_string = pi_string[48:]

    # 4. Gamma = string_to_point(gamma_string)
    gamma = _decode_point(gamma_string)

    # 5. if Gamma = "INVALID" output "INVALID" and stop.
    if gamma == "INVALID":
        return "INVALID"

    # 6. c = string_to_int(c_string)
    c = int.from_bytes(c_string, "little")

    # 7. s = string_to_int(s_string)
    s = int.from_bytes(s_string, "little")

    # 8. Output Gamma, c, and s
    return gamma, c, s


def _assert_and_sample(keys, actuals):
    """
    Input:
        key - key for assert values, basename (+ '_sample') for sampled values.
    Output:
        None; asserts actuals then and assigns into global test_dict
    If key exists, assert dict expected value against provided actual value.
    Sample actual value and store into test_dict under key + '_sample'.
    """
    # noinspection PyGlobalUndefined
    global test_dict
    for key, actual in zip(keys, actuals):
        if key in test_dict and actual:
            assert actual == test_dict[key], "{}  actual:{} != expected:{}".format(
                key, actual.hex(), test_dict[key].hex()
            )
        test_dict[key + "_sample"] = actual


# Much of the following code has been adapted from ed25519.py at https://ed25519.cr.yp.to/software.html retrieved 27 Dec 2019
# While it is gloriously inefficient, it provides an excellent demonstration of the underlying math. For example, production
# code would likely avoid inversion via Fermat's little theorem as it is extremely expensive with a cost of ~300 field multiplies.


def _edwards_add(p, q):
    """Edwards curve point addition"""
    x1 = p[0]
    y1 = p[1]
    x2 = q[0]
    y2 = q[1]
    x3 = (x1 * y2 + x2 * y1) * _inverse(1 + D * x1 * x2 * y1 * y2)
    y3 = (y1 * y2 + x1 * x2) * _inverse(1 - D * x1 * x2 * y1 * y2)
    return [x3 % PRIME, y3 % PRIME]


def _encode_point(p):
    """Encode point to string containing LSB OF X and 254 bits of y"""
    return ((p[1] & ((1 << 255) - 1)) + ((p[0] & 1) << 255)).to_bytes(32, "little")


def _decode_point(s):
    """Decode string containing LSB of X and 254 bits of y into point. Checks on-curve. May return \"INVALID\" """
    y = int.from_bytes(s, "little") & ((1 << 255) - 1)
    x = _x_recover(y)
    if x & 1 != _get_bit(s, BITS - 1):
        x = PRIME - x
    p = [x, y]
    if not _is_on_curve(p):
        return "INVALID"
    return p


def _get_bit(h, i):
    """Return specified bit from string for subsequent testing"""
    h1 = int.from_bytes(h, "little")
    return (h1 >> i) & 0x01


def _get_secret_scalar(sk):
    """Calculate and return the secret_scalar integer"""
    h = bytearray(_hash(sk)[0:32])
    h[31] = int((h[31] & 0x7F) | 0x40)
    h[0] = int(h[0] & 0xF8)
    secret_int = int.from_bytes(h, "little")
    return secret_int


def _hash(message):
    """Return 64-byte SHA512 hash of arbitrary-length byte message"""
    return hashlib.sha512(message).digest()


def _short_hash(message):
    """Return 64-byte SHA512 hash of arbitrary-length byte message... trimmed to 31 byte felt"""
    return hashlib.sha512(message).digest()[:31]


def _inverse(x):
    """Calculate inverse via Fermat's little theorem"""
    return pow(x, PRIME - 2, PRIME)


def _is_on_curve(p):
    """Check to confirm point is on curve; return boolean"""
    x = p[0]
    y = p[1]
    result = (-x * x + y * y - 1 - D * x * x * y * y) % PRIME
    return result == 0


def _scalar_multiply(p, e):
    """Scalar multiplied by curve point"""
    if e == 0:
        return [0, 1]
    q = _scalar_multiply(p, e // 2)
    q = _edwards_add(q, q)
    if e & 1:
        q = _edwards_add(q, p)
    return q


def _x_recover(y):
    """Recover x coordinate from y coordinate"""
    xx = (y * y - 1) * _inverse(D * y * y + 1)
    x = pow(xx, (PRIME + 3) // 8, PRIME)
    if (x * x - xx) % PRIME != 0:
        x = (x * II) % PRIME
    if x % 2 != 0:
        x = PRIME - x
    return x


# Constants, some of which are calculated/checked at runtime using above routines
# See https://ed25519.cr.yp.to/python/checkparams.py
SUITE_STRING = bytes([0x04])
BITS = 256
PRIME = 2**255 - 19
ORDER = 2**252 + 27742317777372353535851937790883648493
COFACTOR = 8
TWO_INV = _inverse(2)
II = pow(2, (PRIME - 1) // 4, PRIME)
A = 486662
D = -121665 * _inverse(121666)
BASEy = 4 * _inverse(5)
BASEx = _x_recover(BASEy)
BASE = [BASEx % PRIME, BASEy % PRIME]
assert BITS >= 10
assert 8 * len(_hash("hash input".encode("UTF-8"))) == 2 * BITS
assert pow(2, PRIME - 1, PRIME) == 1
assert PRIME % 4 == 1
assert pow(2, ORDER - 1, ORDER) == 1
assert ORDER >= 2 ** (BITS - 4)
assert ORDER <= 2 ** (BITS - 3)
assert pow(D, (PRIME - 1) // 2, PRIME) == PRIME - 1
assert pow(II, 2, PRIME) == PRIME - 1
assert _is_on_curve(BASE)
assert _scalar_multiply(BASE, ORDER) == [0, 1]



from enum import IntEnum, unique
from typing import List, Literal

from empiric.core.utils import str_to_felt

ADDRESS = int
HEX_STR = str

# Network Types
STAGING = "staging"
TESTNET = "testnet"
INTEGRATION = "integration"
MAINNET = "mainnet"
SCROLL_TESTNET = "scroll_testnet"
LINEA_TESTNET = "linea_testnet"

Network = Literal["staging", "testnet", "integration", "mainnet"]

CHAIN_IDS = {
    INTEGRATION: 1536727068981429685321,
    TESTNET: 1536727068981429685321,
    MAINNET: 23448594291968334,
    SCROLL_TESTNET: 534352,
    LINEA_TESTNET: 59140,
}

GATEWAY_URLS = {
    TESTNET: "https://alpha4.starknet.io",
    INTEGRATION: "https://external.integration.starknet.io",
    MAINNET: "https://alpha-mainnet.starknet.io",
    SCROLL_TESTNET: "https://scroll-alphanet.public.blastapi.io",
    LINEA_TESTNET: "https://consensys-zkevm-goerli-prealpha.infura.io/v3/ef9b71db32e242f39c6cf0691c8b521a",
}


# aggregation mode enum
@unique
class AggregationMode(IntEnum):
    MEDIAN = 84959893733710  # str_to_felt("MEDIAN")


class Currency:
    id: int
    decimals: int
    is_abstract_currency: int
    starknet_address: int
    ethereum_address: int

    def __init__(
        self,
        id,
        decimals,
        is_abstract_currency,
        starknet_address=None,
        ethereum_address=None,
    ):
        if type(id) == str:
            id = str_to_felt(id)
        self.id = id

        self.decimals = decimals

        if type(is_abstract_currency) == bool:
            is_abstract_currency = int(is_abstract_currency)
        self.is_abstract_currency = is_abstract_currency

        if starknet_address is None:
            starknet_address = 0
        self.starknet_address = starknet_address

        if ethereum_address is None:
            ethereum_address = 0
        self.ethereum_address = ethereum_address

    def serialize(self) -> List[str]:
        return [
            self.id,
            self.decimals,
            self.is_abstract_currency,
            self.starknet_address,
            self.ethereum_address,
        ]

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "decimals": self.decimals,
            "is_abstract_currency": self.is_abstract_currency,
            "starknet_address": self.starknet_address,
            "ethereum_address": self.ethereum_address,
        }


class Pair:
    id: int
    quote_currency_id: int
    base_currency_id: int

    def __init__(self, id, quote_currency_id, base_currency_id):
        if type(id) == str:
            id = str_to_felt(id)
        self.id = id

        if type(quote_currency_id) == str:
            quote_currency_id = str_to_felt(quote_currency_id)
        self.quote_currency_id = quote_currency_id

        if type(base_currency_id) == str:
            base_currency_id = str_to_felt(base_currency_id)
        self.base_currency_id = base_currency_id

    def serialize(self) -> List[str]:
        return [self.id, self.quote_currency_id, self.base_currency_id]

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "quote_currency_id": self.quote_currency_id,
            "base_currency_id": self.base_currency_id,
        }



SUMMARY_STATS_ABI = [
    {
        "inputs": [{"name": "oracle_address", "type": "felt"}],
        "name": "constructor",
        "outputs": [],
        "type": "constructor",
    },
    {
        "inputs": [
            {"name": "key", "type": "felt"},
            {"name": "start", "type": "felt"},
            {"name": "stop", "type": "felt"},
        ],
        "name": "calculate_mean",
        "outputs": [{"name": "mean_", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "key", "type": "felt"},
            {"name": "start", "type": "felt"},
            {"name": "stop", "type": "felt"},
        ],
        "name": "calculate_volatility",
        "outputs": [{"name": "volatility_", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
]



import logging
from typing import Any, Callable, List, Optional

from empiric.core.abis.randomness import RANDOMNESS_ABI
from empiric.core.contract import Contract
from starknet_py.contract import InvokeResult
from starknet_py.net.account.account import Account
from starknet_py.net.client import Client

logger = logging.getLogger(__name__)


class RandomnessMixin:
    client: Client
    account: Account
    randomness: Optional[Contract] = None
    track_nonce: Callable[Any, None] = None

    def init_randomness_contract(
        self,
        randomness_contract_address: int,
    ):
        self.randomness = Contract(
            randomness_contract_address,
            RANDOMNESS_ABI,
            self.client,
        )

    async def request_random(
        self,
        seed: int,
        callback_address: int,
        callback_gas_limit: int = 1000000,
        publish_delay: int = 1,
        num_words: int = 1,
        max_fee=int(1e16),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.randomness.functions["request_random"].invoke(
            seed,
            callback_address,
            callback_gas_limit,
            publish_delay,
            num_words,
            callback=self.track_nonce,
            max_fee=max_fee,
        )
        return invocation

    async def submit_random(
        self,
        request_id: int,
        requestor_address: int,
        seed: int,
        callback_address: int,
        callback_gas_limit: int,  # =1000000
        minimum_block_number: int,
        random_words: List[int],  # List with 1 item
        block_hash: int,  # block hash of block
        proof: List[int],  # randomness proof
        max_fee=int(1e16),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.randomness.functions["submit_random"].invoke(
            request_id,
            requestor_address,
            seed,
            callback_address,
            callback_gas_limit,
            minimum_block_number,
            random_words,
            block_hash,
            proof,
            callback=self.track_nonce,
            max_fee=max_fee,
        )
        return invocation

    async def get_request_status(
        self,
        caller_address: int,
        request_id: int,
    ):
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.randomness.functions["get_request_status"].call(
            caller_address,
            request_id,
        )

        return invocation



import configparser
from typing import Optional

import typer
from empiric.cli import (
    SUCCESS,
    __app_name__,
    account,
    config,
    contracts,
    net,
    publisher,
    randomness,
)
from starkware.crypto.signature.signature import get_random_private_key

from .contracts.utils import deploy_contract
from .utils import coro

# TESTNET_GATEWAY_URL = "https://alpha4.starknet.io"
TESTNET_GATEWAY_URL = "http://127.0.0.1:5050"
app = typer.Typer()


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__}")
        raise typer.Exit()


@app.command()
def init(gateway_url=TESTNET_GATEWAY_URL, chain_id: int = 1536727068981429685321):
    config.init_app(gateway_url, chain_id)
    return SUCCESS


@app.command()
@coro
async def create_account(
    config_path=config.DEFAULT_CONFIG, save_to_config: bool = True
):
    gateway_url, chain_id = config.validate_config(config_path)

    client = net.init_client(gateway_url, chain_id)
    await account.create_account(
        client, config.CONFIG_FILE_PATH, save_to_config=save_to_config
    )
    return SUCCESS


@app.command()
@coro
async def deploy_by_name(contract_name: str, config_path=config.DEFAULT_CONFIG):
    contract_address = await deploy_contract(config_path, contract_name)
    typer.echo(f"address: {contract_address}")
    return SUCCESS


@app.command()
def gen_pvt_key(config_file=config.DEFAULT_CONFIG):
    key = get_random_private_key()

    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)
    try:
        config_parser["SECRET"]["private-key"] = str(key)
    except KeyError:
        config_parser["SECRET"] = {"private-key": str(key)}

    with open(config_file, "w") as f:
        config_parser.write(f)

    return SUCCESS


@app.command()
def local_faucet(config_file=config.DEFAULT_CONFIG):
    """request tokens from the devnet faucet"""
    import requests

    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)
    account_address = int(config_parser["USER"]["address"])

    r = requests.post(
        "http://127.0.0.1:5050/mint",
        json={"address": hex(account_address), "amount": 1e22},
    )
    typer.echo(f"{r.json()}, {r.status_code}")


@app.command()
def get_block(config_path=config.DEFAULT_CONFIG):
    gateway_url, chain_id = config.validate_config(config_path)

    client = net.init_client(gateway_url, chain_id)
    typer.echo(client.get_block_sync())
    return SUCCESS


@app.command()
def devnet():
    """start a local devnet instance"""

    import sys

    import starknet_devnet
    from starknet_devnet.server import main

    sys.argv = [starknet_devnet.__file__]
    sys.exit(main())

    return SUCCESS


@app.command()
def account_address(config_path=config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    typer.echo(client.account_address())
    return SUCCESS


@app.command()
@coro
async def balance_of(account: int, config_path=config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    typer.echo(f"BALANCE: {await client.get_balance(account)}")
    return SUCCESS


@app.command()
def quickstart(
    ctx: typer.Context,
    config_file=config.DEFAULT_CONFIG,
    faucet: Optional[bool] = typer.Option(False, "--faucet", "-f"),
):
    typer.echo("creating config")
    ctx.invoke(init)
    typer.echo("generating private key")
    ctx.invoke(gen_pvt_key, config_file)
    typer.echo("deploying account contract")
    ctx.invoke(create_account, config_file)
    if faucet:
        typer.echo("getting funds from testnet faucet")
        ctx.invoke(local_faucet, config_file)

    return SUCCESS


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the application's version and exit.",
        callback=_version_callback,
        is_eager=True,
    )
) -> None:
    return


app.add_typer(contracts.app, name="contracts")
app.add_typer(publisher.app, name="publisher")
app.add_typer(randomness.app, name="random")



import configparser
from pathlib import Path

from empiric.core import EmpiricClient
from empiric.core.config import ContractAddresses
from starknet_py.net.account.account import Account
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models.chains import StarknetChainId
from starknet_py.net.signer.stark_curve_signer import KeyPair, StarkCurveSigner


def get_gateway_url(config_file: Path) -> str:
    """Return the current url for starknet gateway"""

    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    if config_parser["GENERAL"]["network"] == "testnet":
        return "testnet"
    if config_parser["GENERAL"]["network"] == "mainnet":
        return "mainnet"

    return config_parser["GENERAL"]["gateway-url"]


def get_chain_id(config_file: Path) -> str:
    """Return the current url for starknet gateway"""

    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    return int(config_parser["GENERAL"]["chain-id"])


def init_client(gateway_url: str, chain_id: int) -> int:
    return GatewayClient(gateway_url, chain=StarknetChainId(chain_id))


def init_empiric_client(config_file: Path) -> EmpiricClient:
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    account_private_key = int(config_parser["SECRET"]["private-key"])
    network = config_parser["GENERAL"]["network"]
    account_contract_address = int(config_parser["USER"]["address"])
    publisher_registry_address = int(
        config_parser["CONTRACTS"].get("publisher-registry", 0)
    )
    oracle_proxy_address = int(config_parser["CONTRACTS"].get("oracle-proxy", 0))

    client = EmpiricClient(
        network,
        account_private_key=account_private_key,
        account_contract_address=account_contract_address,
        contract_addresses_config=ContractAddresses(
            publisher_registry_address, oracle_proxy_address
        ),
    )
    return client


def init_account_client(client: GatewayClient, config_file: Path) -> Account:
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    account_private_key = int(config_parser["SECRET"]["private-key"])
    account_contract_address = int(config_parser["USER"]["address"])
    chain_id = int(config_parser["GENERAL"]["chain-id"])

    key_pair = KeyPair.from_private_key(account_private_key)
    signer = StarkCurveSigner(
        account_address=account_contract_address,
        key_pair=key_pair,
        chain_id=StarknetChainId(chain_id),
    )

    return Account(
        address=account_contract_address,
        client=client,
        signer=signer,
    )



import configparser
import json
import time
from pathlib import Path
from typing import Optional, Tuple

import typer
from empiric.cli import SUCCESS, config, net
from empiric.cli.contracts.utils import (
    DEFAULT_MAX_FEE,
    _format_currencies,
    _format_pairs,
)
from empiric.cli.utils import coro
from empiric.core.utils import str_to_felt
from starknet_py.contract import Contract
from starknet_py.net.client import Client
from starkware.starknet.compiler.compile import get_selector_from_name

from .utils import declare_contract

app = typer.Typer(help="Deployment commands for Oracle")
ORACLE_CONFIG = typer.Option(
    "",
    "--deploy-config",
    "-d",
    help="configuration for currency and pair deployment",
)


@app.command()
@coro
async def deploy(
    cli_config=config.DEFAULT_CONFIG, deploy_config: Optional[str] = ORACLE_CONFIG
):
    """
    Deploy a new proxied instance of the publisher registry.
    This requires a configuration file for the currencies and pairs that the oracle will support.
    There is a sample config called oracle_constructor_data.json that shows the format.

    """
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)

    deploy_config = deploy_config or config_parser["CONFIG"]["oracle-config-path"]

    # TODO (rlkelly): allow setting default path for config lookup in cli config
    deploy_config_path = Path(deploy_config)
    if not deploy_config_path.is_file():
        typer.echo(
            "No valid config path, please provide arg using --deploy_config or create a copy of "
            "cli/sample_config/oracle_constructor_data.json in the current path"
        )
        return 1

    gateway_url, chain_id = config.validate_config(cli_config)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, cli_config)

    await deploy_oracle_proxy(account_client, deploy_config_path, cli_config)

    return SUCCESS


@app.command()
@coro
async def upgrade(cli_config=config.DEFAULT_CONFIG):
    valid = input("Are you sure?\nThis will update the contract. (y/n)\n")
    if valid != "y":
        return SUCCESS
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )

    client = net.init_empiric_client(cli_config)
    declared_oracle_class_hash = await declare_contract(
        client.client, compiled_contract_path, "Oracle"
    )
    typer.echo(f"DECLARED: {declared_oracle_class_hash}")
    invocation = await client.oracle.upgrade.invoke(
        declared_oracle_class_hash, max_fee=DEFAULT_MAX_FEE
    )
    await invocation.wait_for_acceptance()
    typer.echo(f"INVOKED: {invocation.hash}")

    return SUCCESS


@app.command()
@coro
async def publish_spot_entry(entry: str, config_path=config.DEFAULT_CONFIG):
    pair_id, value, timestamp, source, publisher = entry.split(",")
    if timestamp.lower() == "now":
        timestamp = int(time.time())

    await _publish_spot_entry(
        config_path,
        (
            str_to_felt(pair_id.lower()),
            int(value),
            int(timestamp),
            str_to_felt(source),
            str_to_felt(publisher),
        ),
    )

    return SUCCESS


@app.command()
@coro
async def cp(pair_id: str, config_path=config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    invocation = await client.oracle.set_checkpoint.invoke(
        str_to_felt(pair_id),
        0,
        max_fee=DEFAULT_MAX_FEE,
    )
    await invocation.wait_for_acceptance(wait_for_accept=True)
    typer.echo(f"invocation: {invocation.hash}")

    return SUCCESS


@app.command()
@coro
async def get_cp(pair_id: str, index: int, config_path=config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    invocation = await client.oracle.get_checkpoint.call(
        str_to_felt(pair_id),
        index,
    )
    typer.echo(f"invocation: {invocation}")

    return SUCCESS


async def deploy_oracle_proxy(
    client: Client, deploy_config_path: Path, config_path: Path
):
    """starknet deploy --contract contracts/starknet/build/PublisherRegistry.json --inputs <ADMIN_ADDRESS>"""
    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )

    deploy_config = json.loads(deploy_config_path.read_text("utf-8"))
    currencies = deploy_config["currencies"]
    pairs = deploy_config["pairs"]

    admin_address = int(config_parser["USER"]["address"])
    publisher_registry_address = int(config_parser["CONTRACTS"]["publisher-registry"])

    declared_oracle_class_hash = await declare_contract(
        client, compiled_contract_path, "Oracle"
    )
    compiled_proxy = (compiled_contract_path / "Proxy.json").read_text("utf-8")
    declare_result = await Contract.declare(
        client, compiled_contract=compiled_proxy, max_fee=int(1e16)
    )
    await declare_result.wait_for_acceptance()
    deployment_result = await declare_result.deploy(
        constructor_args=[
            declared_oracle_class_hash,
            get_selector_from_name("initializer"),
            [
                admin_address,
                publisher_registry_address,
                len(currencies),
                *_format_currencies(currencies),
                len(pairs),
                *_format_pairs(pairs),
            ],
        ],
        max_fee=int(1e16),
    )
    await deployment_result.wait_for_acceptance()
    typer.echo(f"proxy address: {deployment_result.deployed_contract.address}")

    oracle_proxy_address = deployment_result.deployed_contract.address
    config_parser["CONTRACTS"]["oracle-proxy"] = str(oracle_proxy_address)

    with open(config_path, "w") as f:
        config_parser.write(f)


async def _publish_spot_entry(config_path: Path, entry: Tuple[int, int, int, int, int]):
    client = net.init_empiric_client(config_path)
    invocation = await client.publish_spot_entry(*entry)

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def get_spot(pair_id: str, config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    entry = await client.oracle.get_spot.call(str_to_felt(pair_id), 0)
    typer.echo(f"spot: {entry}")


@app.command()
@coro
async def get_latest_cp(pair_id: str, config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    latest = await client.oracle.get_latest_checkpoint_index.call(str_to_felt(pair_id))
    typer.echo(f"latest: {latest}")
    entry = await client.oracle.get_last_spot_checkpoint_before.call(
        str_to_felt(pair_id), 1666197238
    )
    typer.echo(f"cp: {entry}")


@app.command()
@coro
async def get_spot_for_sources(pair_id: str, config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    entry = await client.oracle.get_spot_for_sources.call(
        str_to_felt(pair_id), 0, ["CEX"]
    )
    typer.echo(f"spot: {entry}")


@app.command()
@coro
async def get_spot_entries(pair_id: str, config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    entries = await client.oracle.get_spot_entries.call(str_to_felt(pair_id))
    typer.echo(f"spot: {entries}")


@app.command()
@coro
async def get_spot_entry(
    pair_id: str, config_path: Path = config.DEFAULT_CONFIG, publisher: str = "CEX"
):
    client = net.init_empiric_client(config_path)
    entry = await client.oracle.get_spot_entry.call(
        str_to_felt(pair_id), str_to_felt(publisher)
    )
    typer.echo(f"spot: {entry}")


@app.command()
@coro
async def get_spot_decimals(pair_id: str, config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    entry = await client.oracle.get_spot_decimals.call(str_to_felt(pair_id))
    typer.echo(f"decimals: {entry}")



import configparser
from pathlib import Path

import typer
from empiric.cli import SUCCESS
from starknet_py.contract import Contract
from starknet_py.net import KeyPair
from starknet_py.net.client import Client
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models.address import AddressRepresentation

from .compiled_account_contract import COMPILED_ACCOUNT_CONTRACT


async def deploy_account_contract(
    client: Client, public_key: int
) -> AddressRepresentation:
    declare_result = await Contract.declare(
        account=client, compiled_contract=COMPILED_ACCOUNT_CONTRACT, max_fee=int(1e16)
    )
    await client.wait_for_tx(
        tx_hash=declare_result.hash,
        wait_for_accept=True,
    )
    deploy_result = await declare_result.deploy(
        max_fee=int(1e16), constructor_calldata=[public_key]
    )
    await client.wait_for_tx(
        tx_hash=deploy_result.hash,
        wait_for_accept=True,
    )
    return deploy_result.deployed_contract.address


async def create_account(
    client: GatewayClient, config_file: Path, save_to_config: bool = True
):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    account_private_key = int(config_parser["SECRET"]["private-key"])

    key_pair = KeyPair.from_private_key(account_private_key)
    address = await deploy_account_contract(client, key_pair.public_key)
    typer.echo(f"created address: {address}")

    if save_to_config:
        config_parser["USER"]["address"] = str(address)
        with open(config_file, "w") as f:
            config_parser.write(f)

    return SUCCESS



from .bitstamp import BitstampFetcher
from .cex import CexFetcher
from .kaiko import KaikoFetcher
from .coinbase import CoinbaseFetcher
from .ascendex import AscendexFetcher
from .okx import OkxFetcher
from .coingecko import CoingeckoFetcher
from .gemini import GeminiFetcher
from .thegraph import TheGraphFetcher
from .defillama import DefillamaFetcher



__app_name__ = "empiric_cli"

(
    SUCCESS,
    STARKNET_READ_ERROR,
    STARKNET_WRITE_ERROR,
) = range(3)

(
    DIR_ERROR,
    FILE_ERROR,
    OS_ERROR,
) = range(3)

ERRORS = {
    STARKNET_READ_ERROR: "starknet read error",
    STARKNET_WRITE_ERROR: "starknet write error",
    DIR_ERROR: "DIRECTORY ERROR",
    OS_ERROR: "OS ERROR",
}



import configparser
import os
from pathlib import Path

import typer
from empiric.cli import DIR_ERROR, FILE_ERROR, OS_ERROR, SUCCESS, net

CONFIG_DIR_PATH = Path(os.getcwd())
COMPILED_CONTRACT_PATH = Path(os.getcwd()) / "contracts" / "starknet" / "build"
CONFIG_FILE_PATH = CONFIG_DIR_PATH / "cli-config.ini"
DEFAULT_CONFIG = typer.Option(
    CONFIG_FILE_PATH, "--config-path", "-c", help="optional path to cli config"
)


def init_app(gateway_url: str, chain_id: int) -> int:
    """Initialize the application."""
    config_code = _init_config_file()
    if config_code != SUCCESS:
        return config_code
    _create_config_file(gateway_url, chain_id)

    return SUCCESS


def validate_config(config_path):
    if type(config_path) == str:
        config_path = Path(config_path)
    if config_path.exists():
        gateway_url = net.get_gateway_url(config_path)
        chain_id = net.get_chain_id(config_path)
    else:
        typer.secho(
            'Config file not found. Please, run "python3 -m cli init"',
            fg=typer.colors.RED,
        )
        raise typer.Exit(1)
    return gateway_url, chain_id


def _init_config_file() -> int:
    try:
        CONFIG_DIR_PATH.mkdir(exist_ok=True)
    except OSError:
        return DIR_ERROR

    try:
        CONFIG_FILE_PATH.touch(exist_ok=True)
    except OSError:
        return FILE_ERROR

    return SUCCESS


def _create_config_file(gateway_url: str, chain_id: int) -> int:
    config_parser = configparser.ConfigParser()
    config_parser["GENERAL"] = {
        "network": "local",
        "gateway-url": gateway_url,
        "chain-id": chain_id,
    }
    config_parser["SECRET"] = {}
    config_parser["USER"] = {}
    config_parser["CONTRACTS"] = {}
    config_parser["CONFIG"] = {
        "contract-path": str(COMPILED_CONTRACT_PATH),
        "oracle-config-path": "./oracle_constructor_data.json",
    }

    try:
        with CONFIG_FILE_PATH.open("w") as file:
            config_parser.write(file)

    except OSError:
        return OS_ERROR

    return SUCCESS



import configparser
import time
from pathlib import Path

import typer
from empiric.cli import SUCCESS, config, net
from empiric.cli.utils import coro
from starknet_py.contract import Contract
from starknet_py.net.client import Client

app = typer.Typer(help="Deployment commands for SummaryStats")


@app.command()
@coro
async def deploy(cli_config=config.DEFAULT_CONFIG):
    """
    Deploy a new proxied instance of the publisher registry.
    This requires a configuration file for the currencies and pairs that the oracle will support.
    There is a sample config called oracle_constructor_data.json that shows the format.

    """
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)

    gateway_url, chain_id = config.validate_config(cli_config)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, cli_config)

    await _deploy_summary_stats(account_client, cli_config)

    return SUCCESS


async def _deploy_summary_stats(client: Client, config_path: Path):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    oracle_proxy_address = int(config_parser["CONTRACTS"]["oracle-proxy"])

    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )
    compiled_contract = (compiled_contract_path / "SummaryStats.json").read_text(
        "utf-8"
    )
    declare_result = await Contract.declare(
        client,
        compiled_contract=compiled_contract,
    )
    await declare_result.wait_for_acceptance()
    deployment_result = await declare_result.deploy(
        constructor_args=[
            oracle_proxy_address,
        ],
    )
    await deployment_result.wait_for_acceptance()
    typer.echo(f"summary stats address: {deployment_result.deployed_contract.address}")

    summary_stats_address = deployment_result.deployed_contract.address
    config_parser["CONTRACTS"]["summary-stats"] = str(summary_stats_address)

    with open(config_path, "w") as f:
        config_parser.write(f)


@app.command()
@coro
async def volatility(
    pair_id: str,
    start: int = int(time.time() - 3600),
    end: int = int(time.time()),
    config_path: Path = config.DEFAULT_CONFIG,
):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)

    client = net.init_empiric_client(config_path)
    client.init_stats_contract(int(config_parser["CONTRACTS"]["summary-stats"]))

    response = await client.stats.calculate_volatility.call(pair_id, start, end)
    typer.echo(f"VOLATILITY: {response}")

    return SUCCESS



import asyncio
from typing import List

import aiohttp
from empiric.core.client import EmpiricClient
from empiric.core.entry import SpotEntry
from empiric.publisher.types import PublisherInterfaceT


class EmpiricPublisherClient(EmpiricClient):
    """
    This client extends the empiric client with functionality for fetching from our third party sources.
    It can be used to synchronously or asynchronously fetch assets using the Asset format, ie.

    `{"type": "SPOT", "pair": ("BTC", "USD"), "decimals": 18}`

    More to follow on the standardization of this format.

    The client works by setting up fetchers that are provided the assets to fetch and the publisher name.

    ```python
    cex_fetcher = CexFetcher(EMPIRIC_ALL_ASSETS, "empiric_fetcher_test")
    gemini_fetcher = GeminiFetcher(EMPIRIC_ALL_ASSETS, "empiric_fetcher_test")
    fetchers = [
        cex_fetcher,
        gemini_fetcher,
    ]
    eapc = EmpiricPublisherClient('testnet')
    eapc.add_fetchers(fetchers)
    await eapc.fetch()
    eapc.fetch_sync()
    ```
    """

    fetchers: List[PublisherInterfaceT] = []

    @staticmethod
    def convert_to_publisher(client: EmpiricClient):
        client.__class__ = EmpiricPublisherClient
        return client

    def add_fetchers(self, fetchers: List[PublisherInterfaceT]):
        self.fetchers.extend(fetchers)

    def add_fetcher(self, fetcher: PublisherInterfaceT):
        self.fetchers.append(fetcher)

    async def fetch(self, filter_exceptions=True) -> List[SpotEntry]:
        tasks = []
        timeout = aiohttp.ClientTimeout(total=10)  # 10 seconds per request
        async with aiohttp.ClientSession(timeout=timeout) as session:
            for fetcher in self.fetchers:
                data = fetcher.fetch(session)
                tasks.append(data)
            result = await asyncio.gather(*tasks, return_exceptions=True)
            if filter_exceptions:
                result = [subl for subl in result if not isinstance(subl, Exception)]
            return [val for subl in result for val in subl]

    def fetch_sync(self) -> List[SpotEntry]:
        results = []
        for fetcher in self.fetchers:
            data = fetcher.fetch_sync()
            results.extend(data)
        return results



import configparser
from pathlib import Path
from typing import List

import typer
from empiric.cli import SUCCESS, config, net
from empiric.cli.utils import coro
from empiric.core.utils import felt_to_str, str_to_felt
from starknet_py.contract import Contract
from starknet_py.net.gateway_client import GatewayClient

app = typer.Typer(help="Deployment commands for Publisher Registry")


@app.command()
@coro
async def deploy(config_path=config.DEFAULT_CONFIG):
    """deploy a new instance of the publisher registry"""
    gateway_url, chain_id = config.validate_config(config_path)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, config_path)

    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )

    await deploy_publisher_registry(account_client, compiled_contract_path, config_path)

    return SUCCESS


@app.command()
@coro
async def add_publisher(
    publisher, publisher_address: int, config_path=config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    invocation = await client.publisher_registry.add_publisher.invoke(
        str_to_felt(publisher), publisher_address, max_fee=int(1e16)
    )

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def update_publisher(
    publisher, publisher_address: int, config_path=config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    invocation = await client.publisher_registry.update_publisher_address.invoke(
        str_to_felt(publisher), publisher_address, max_fee=int(1e16)
    )

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def register_self(publisher: str, config_path=config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    publisher_address = client.account_address()
    invocation = await client.add_publisher(publisher, publisher_address)

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def add_source_for_publisher(
    publisher: str, source: str, config_path=config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    invocation = await client.add_source_for_publisher(publisher, source)

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def add_sources_for_publisher(
    publisher: str, sources: List[str], config_path=config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    invocation = await client.add_sources_for_publisher(publisher, sources)

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def update_publisher_address(
    publisher: str, new_address: int, config_path=config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    invocation = await client.update_publisher_address(publisher, new_address)

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def get_all_publishers(config_path: Path = config.DEFAULT_CONFIG):
    client = net.init_empiric_client(config_path)
    publishers = await client.publisher_registry.get_all_publishers.call()
    typer.echo(f"publishers: {[felt_to_str(p) for p in publishers[0]]}")


async def deploy_publisher_registry(
    client: GatewayClient, compiled_contract_path: Path, config_path: Path
):
    """starknet deploy --contract contracts/starknet/build/PublisherRegistry.json --inputs <ADMIN_ADDRESS>"""
    compiled = (compiled_contract_path / "PublisherRegistry.json").read_text("utf-8")

    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)

    admin_address = int(config_parser["USER"]["address"])

    declare_result = await Contract.declare(
        client, compiled_contract=compiled, max_fee=int(1e16)
    )
    await declare_result.wait_for_acceptance()
    deployment_result = await declare_result.deploy(
        constructor_args={"admin_address": admin_address}, max_fee=int(1e16)
    )
    await deployment_result.wait_for_acceptance()
    typer.echo(f"address: {deployment_result.deployed_contract.address}")

    publisher_registry_address = deployment_result.deployed_contract.address
    config_parser["CONTRACTS"]["publisher-registry"] = str(publisher_registry_address)

    with open(config_path, "w") as f:
        config_parser.write(f)


@app.command()
@coro
async def get_metadata(
    publisher: str = "EMPIRIC", config_path: Path = config.DEFAULT_CONFIG
):
    client = net.init_empiric_client(config_path)
    publishers = await client.publisher_registry.get_publisher_sources.call(
        str_to_felt(publisher)
    )
    z = await client.publisher_registry.get_publisher_address.call(
        str_to_felt(publisher)
    )
    typer.echo(f"sources: {[felt_to_str(p) for p in publishers[0]]}")
    typer.echo(f"publisher_address: {z}")



from empiric.core.mixins.nonce import NonceMixin
from empiric.core.mixins.oracle import OracleMixin
from empiric.core.mixins.publisher_registry import PublisherRegistryMixin
from empiric.core.mixins.randomness import RandomnessMixin
from empiric.core.mixins.transactions import TransactionMixin



import asyncio
from functools import wraps

from empiric.cli.contracts.utils import *  # noqa: F401, F403


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper



import typer
from empiric.cli import config, net
from empiric.cli.contracts import oracle, publisher_registry, summary_stats
from empiric.cli.utils import coro

app = typer.Typer(help="contract deployment utilities")

app.add_typer(
    publisher_registry.app,
    name="publisher-registry",
    help="publisher registry contract utils",
)
app.add_typer(
    publisher_registry.app,
    name="pr",
    help="publisher registry contract utils",
)
app.add_typer(oracle.app, name="oracle", help="Oracle contract utils")
app.add_typer(
    summary_stats.app, name="summary_stats", help="SummaryStats contract utils"
)


@app.command()
def list():
    """List all implemented contracts"""
    typer.echo("publisher_registry")
    typer.echo("oracle")


@app.command()
def deploy(
    ctx: typer.Context,
    config_file: str = config.DEFAULT_CONFIG,
    oracle_config: str = oracle.ORACLE_CONFIG,
):
    """List all implemented contracts"""
    typer.echo("deploy Publisher Registry")
    ctx.invoke(publisher_registry.deploy, config_file)
    typer.echo("oracle")
    ctx.invoke(oracle.deploy, config_file, oracle_config)



import configparser
import sys
import time
import traceback
from pathlib import Path

import requests
import typer
from empiric.cli import SUCCESS, config, net
from empiric.cli.contracts.utils import declare_contract
from empiric.cli.randomness.utils import (
    create_randomness,
    ecvrf_verify,
    felt_to_secret_key,
    get_blockhash,
    get_blocknumber,
    get_events,
    uint256_to_2_128,
    verify_randomness,
)
from empiric.cli.utils import coro
from starknet_py.contract import Contract
from starknet_py.net.client import Client
from starknet_py.transaction_exceptions import TransactionRejectedError
from starkware.starknet.compiler.compile import get_selector_from_name

app = typer.Typer(help="randomness utilities")
RANDOMNESS_CONFIG = typer.Option(
    "",
    "--deploy-config",
    "-d",
    help="configuration for randomness",
)


@app.command()
@coro
async def deploy(cli_config=config.DEFAULT_CONFIG):
    """
    Deploy a new proxied instance of the randomness contract
    """
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)

    gateway_url, chain_id = config.validate_config(cli_config)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, cli_config)

    await deploy_randomness_proxy(account_client, cli_config)

    return SUCCESS


@app.command()
@coro
async def deploy_tester(cli_config=config.DEFAULT_CONFIG):
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)

    gateway_url, chain_id = config.validate_config(cli_config)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, cli_config)

    await _deploy_tester(account_client, cli_config)

    return SUCCESS


async def deploy_randomness_proxy(client: Client, config_path: Path):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )

    admin_address = int(config_parser["USER"]["address"])

    declared_randomness_class_hash = await declare_contract(
        client, compiled_contract_path, "Randomness"
    )
    compiled_proxy = (compiled_contract_path / "Proxy.json").read_text("utf-8")
    account_private_key = int(config_parser["SECRET"]["private-key"])
    vrf_private_key = felt_to_secret_key(account_private_key)
    vrf_pvt_int = int.from_bytes(vrf_private_key, sys.byteorder)
    vrf_uint256 = uint256_to_2_128(vrf_pvt_int)
    declare_result = await Contract.declare(
        client,
        compiled_contract=compiled_proxy,
    )
    await declare_result.wait_for_acceptance()
    deployment_result = await declare_result.deploy(
        constructor_args=[
            declared_randomness_class_hash,
            get_selector_from_name("initializer"),
            [
                admin_address,
                vrf_uint256[0],
                vrf_uint256[1],  # Uint256
            ],
        ],
    )
    await deployment_result.wait_for_acceptance()
    typer.echo(f"proxy address: {deployment_result.deployed_contract.address}")

    randomness_proxy_address = deployment_result.deployed_contract.address
    config_parser["CONTRACTS"]["randomness-proxy"] = str(randomness_proxy_address)

    with open(config_path, "w") as f:
        config_parser.write(f)


@app.command()
@coro
async def upgrade(cli_config=config.DEFAULT_CONFIG):
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )
    client = net.init_empiric_client(cli_config)

    declared_randomness_class_hash = await declare_contract(
        client.client, compiled_contract_path, "Randomness"
    )

    randomness_contract_address = int(config_parser["CONTRACTS"]["randomness-proxy"])
    client.init_randomness_contract(randomness_contract_address)

    invocation = await client.randomness.functions["upgrade"].invoke(
        declared_randomness_class_hash,
        max_fee=int(1e18),
    )
    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


async def _deploy_tester(client: Client, config_path: Path):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )
    compiled_example = (compiled_contract_path / "ExampleRandomness.json").read_text(
        "utf-8"
    )

    declare_result = await Contract.declare(
        client,
        compiled_contract=compiled_example,
    )
    await declare_result.wait_for_acceptance()

    deployment_result = await declare_result.deploy(
        constructor_args=[],
    )
    await deployment_result.wait_for_acceptance()
    typer.echo(f"tester address: {deployment_result.deployed_contract.address}")

    example_address = deployment_result.deployed_contract.address
    config_parser["CONTRACTS"]["example-randomness"] = str(example_address)

    with open(config_path, "w") as f:
        config_parser.write(f)


@app.command()
@coro
async def request_random(
    seed: int, callback_address: int, cli_config=config.DEFAULT_CONFIG
):
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    randomness_contract_address = int(config_parser["CONTRACTS"]["randomness-proxy"])

    client = net.init_empiric_client(cli_config)
    client.init_randomness_contract(randomness_contract_address)

    invocation = await client.request_random(
        seed=seed,
        callback_address=callback_address,
    )
    typer.echo(f"hash: {invocation.hash}")
    response = await invocation.wait_for_acceptance()

    typer.echo(f"response hash: {response.hash}")


@app.command()
@coro
async def submit_random(
    request_id: int,
    requestor_address: int,
    seed: int,
    callback_address: int,
    callback_gas_limit: int,  # =1000000
    minimum_block_number: int,
    random_words: str,  # List with 1 item
    block_hash: int,  # block hash of block
    proof: str,  # randomness proof
    cli_config=config.DEFAULT_CONFIG,
):
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    randomness_contract_address = int(config_parser["CONTRACTS"]["randomness-proxy"])

    random_words = random_words.split(",")
    proof = proof.split(",")

    client = net.init_empiric_client(cli_config)
    client.init_randomness_contract(randomness_contract_address)

    invocation = await client.submit_random(
        request_id,
        requestor_address,
        seed,
        callback_address,
        callback_gas_limit,
        minimum_block_number,
        random_words,
        block_hash,
        proof,
    )

    await invocation.wait_for_acceptance()
    typer.echo(f"response hash: {invocation.hash}")


@app.command()
@coro
async def handle_random(min_block: int = 0, cli_config=config.DEFAULT_CONFIG):
    # TODO (rlkelly): this is hardcoded for testnet currently
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    randomness_contract_address = int(config_parser["CONTRACTS"]["randomness-proxy"])
    node_url = config_parser["SECRET"]["node-url"]
    account_private_key = int(config_parser["SECRET"]["private-key"])

    client = net.init_empiric_client(cli_config)
    client.init_randomness_contract(randomness_contract_address)

    block_number = await get_blocknumber(node_url)
    sk = felt_to_secret_key(account_private_key)

    continuation_token = "0"

    while continuation_token:
        event_list = get_events(
            hex(randomness_contract_address), node_url, min_block, continuation_token
        )
        continuation_token = event_list["continuation_token"]

        for event in event_list["events"]:
            if event.minimum_block_number > block_number:
                continue
            status = await client.get_request_status(
                event.caller_address, event.request_id
            )
            if status.status_ != 1:
                continue

            typer.echo(f"event {event}")

            block_hash = await get_blockhash(event.minimum_block_number, node_url)

            seed = (
                event.request_id.to_bytes(8, sys.byteorder)
                + block_hash.to_bytes(32, sys.byteorder)
                + event.seed.to_bytes(32, sys.byteorder)
                + event.caller_address.to_bytes(32, sys.byteorder)
            )
            beta_string, pi_string, _pub = create_randomness(sk, seed)
            beta_string = int.from_bytes(beta_string, sys.byteorder)
            proof = [
                int.from_bytes(p, sys.byteorder)
                for p in [pi_string[:31], pi_string[31:62], pi_string[62:]]
            ]
            random_words = [beta_string]

            try:
                invocation = await client.submit_random(
                    event.request_id,
                    event.caller_address,
                    event.seed,
                    event.minimum_block_number,
                    event.callback_address,
                    event.callback_gas_limit,
                    random_words,
                    block_hash,
                    proof,
                )
            except TransactionRejectedError:
                typer.echo(f"Error submitting randomness for {event}")
                typer.echo(traceback.format_exc())
                await client.update_nonce_dict()
                client.cleanup_nonce_dict()
                continue

            typer.echo(f"submitted tx: {hex(invocation.hash)}\n\n")

        while client.nonce_status:
            typer.echo(str(client.nonce_status))
            typer.echo("...\n\n")
            await client.update_nonce_dict()
            client.cleanup_nonce_dict()
            time.sleep(10)


@app.command()
@coro
async def verify_random(transaction_hash: str, cli_config=config.DEFAULT_CONFIG):
    """provide the hex transaction number to verify the proof for that transaction.  If no event is found will alert the user"""
    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)
    node_url = config_parser["SECRET"]["node-url"]
    # TODO (rlkelly): fetch public key from vrf
    pub_key = b"\xeb\xc1~\xdb\xe9\x00\x9cJ,\xe9\xb1Z`\xee\xe9\xc5\xaf\xb9\xa4\x19+\xd5\x1b6F\x00`\x19\x86\xc3\x1e\xfe"

    url = "https://starknet-archive.hasura.app/v1/graphql"
    request_json = {
        "query": (
            """
                query {
                    event(
                        where: {name:{_eq:"Randomness__proof"}, transaction_hash: {_eq: "{}"}}
                    ) {
                        name
                        arguments {
                        name
                        value
                        }
                    }
                }
            """.replace(
                "{}", transaction_hash
            )
        ),
    }
    r = requests.post(url=url, json=request_json)
    event = r.json()["data"]["event"][0]["arguments"]
    event_dict = {e["name"]: e["value"] for e in event}
    block_hash = await get_blockhash(
        int(event_dict["minimum_block_number"], 16), node_url
    )
    seed = int(event_dict["seed"], 16)
    request_id = int(event_dict["request_id"], 16)
    requestor_address = int(event_dict["requestor_address"], 16)

    seed = (
        request_id.to_bytes(8, sys.byteorder)
        + block_hash.to_bytes(32, sys.byteorder)
        + seed.to_bytes(32, sys.byteorder)
        + requestor_address.to_bytes(32, sys.byteorder)
    )
    proof_ = event_dict["proof"]
    p0 = int(proof_[0], 16).to_bytes(31, sys.byteorder)
    p1 = int(proof_[1], 16).to_bytes(31, sys.byteorder)
    p2 = int(proof_[2], 16).to_bytes(18, sys.byteorder)
    _proof = p0 + p1 + p2

    status, val = ecvrf_verify(pub_key, _proof, seed)
    random_word = int(event_dict["random_words"][0], 16)

    typer.echo(f"status: {status}")
    typer.echo(f"verified random value: {int.from_bytes(val, sys.byteorder)}")
    typer.echo(f"onchain random value:  {random_word}")



import secrets
import sys
from typing import List, Union

import requests
from starknet_py.net.networks import TESTNET

from .randomness_utils import (
    ecvrf_proof_to_hash,
    ecvrf_prove,
    ecvrf_verify,
    get_public_key,
)


class RandomnessRequest:
    def __init__(
        self,
        request_id,
        caller_address,
        seed,
        minimum_block_number,
        callback_address,
        callback_gas_limit,
        num_words,
    ):
        self.request_id = int(request_id, 16)
        self.caller_address = int(caller_address, 16)
        self.seed = int(seed, 16)
        self.minimum_block_number = int(minimum_block_number, 16)
        self.callback_address = int(callback_address, 16)
        self.callback_gas_limit = int(callback_gas_limit, 16)
        self.num_words = int(num_words, 16)

    def __repr__(self):
        return (
            f"Request(caller_address={self.caller_address}, request_id={self.request_id}, "
            f"minimum_block_number={self.minimum_block_number}, callback_address={self.callback_address}, "
            f"seed={self.seed}, minimum_block_number={self.minimum_block_number}, "
            f"callback_gas_limit={self.callback_gas_limit}, num_words={self.num_words})"
        )


async def get_blockhash(block_number, node_url, network=TESTNET):
    r = requests.post(
        node_url,
        json={
            "jsonrpc": "2.0",
            "id": "0",
            "method": "starknet_getBlockWithTxs",
            "params": [{"block_number": block_number}],
        },
    )
    return int(r.json()["result"]["block_hash"], 16)


async def get_blocknumber(node_url, network=TESTNET):
    r = requests.post(
        node_url,
        json={
            "jsonrpc": "2.0",
            "id": "0",
            "method": "starknet_blockNumber",
            "params": [],
        },
    )
    return r.json()["result"]


def get_events(
    contract_address: str,
    node_url,
    min_block: int = 0,
    continuation_token: Union[int, None] = None,
    keys: List[str] = [
        "0xc285ec4fd3baa2fd5b1dc432a00bd5301d2c84b86a7e6900c13b6634b4e81a"
    ],
):
    params = {
        "jsonrpc": "2.0",
        "id": "0",
        "method": "starknet_getEvents",
        "params": [
            {
                "address": contract_address,
                "chunk_size": 30,
                "continuation_token": continuation_token,
                "keys": keys,
                "from_block": {"block_number": 352929},
            }
        ],
    }

    r = requests.post(node_url, json=params).json()
    continuation_token = r["result"].get("continuation_token")  # noqa: F841

    return {
        "events": [RandomnessRequest(*r["data"]) for r in r["result"]["events"]],
        "continuation_token": continuation_token,
    }


def make_secret_key():
    return secrets.token_bytes(nbytes=32)


def felt_to_secret_key(sk):
    return sk.to_bytes(32, sys.byteorder)


def uint256_to_2_128(num: int):
    num = num.to_bytes(32, sys.byteorder)
    return (
        int.from_bytes(num[:16], sys.byteorder),
        int.from_bytes(num[16:], sys.byteorder),
    )


def create_randomness(
    secret_key,
    seed: int,
):
    # Alice generates a secret and public key pair
    public_key = get_public_key(secret_key)

    p_status, pi_string = ecvrf_prove(secret_key, seed)
    b_status, beta_string = ecvrf_proof_to_hash(pi_string)
    assert b_status == "VALID"

    return beta_string, pi_string, public_key


def verify_randomness(
    public_key,
    proof_,
    seed: int,
):
    seed = seed.to_bytes(32, sys.byteorder)[:32]
    result, beta_string2 = ecvrf_verify(public_key, proof_, seed)
    return result



import json
from pathlib import Path
from typing import Dict, List, Union

from empiric.cli import config
from empiric.core.utils import str_to_felt
from starknet_py.common import create_compiled_contract
from starknet_py.contract import Contract
from starknet_py.net.client import Client
from starknet_py.transactions.declare import make_declare_tx
from starknet_py.utils.crypto.facade import pedersen_hash
from starkware.starknet.core.os.class_hash import compute_class_hash

DEFAULT_MAX_FEE = int(1e18)


async def declare_contract(
    client: Client, compiled_contract_path: str, contract_name: str
) -> str:
    compiled_contract = (
        Path(compiled_contract_path) / f"{contract_name}.json"
    ).read_text("utf-8")
    declare_tx = make_declare_tx(compiled_contract=compiled_contract)

    await client.declare(declare_tx)
    compiled_ = create_compiled_contract(None, compiled_contract, None)
    declared_class_hash = compute_class_hash(compiled_, hash_func=pedersen_hash)
    return declared_class_hash


def get_contract(
    contract_address: int,
    contract_name: str,
    client: Client,
    compiled_contract_path: Path = config.COMPILED_CONTRACT_PATH,
):
    abi = json.loads(
        (compiled_contract_path / f"{contract_name}_abi.json").read_text("utf-8")
    )
    return Contract(
        address=contract_address,
        abi=abi,
        provider=client,
    )


def _format_currencies(currencies: Dict[str, str]) -> List[str]:
    # TODO (rlkelly): use marshmallow to format
    output = []
    for row in currencies:
        if isinstance(row["id"], str):
            output.append(str_to_felt(row["id"]))
        else:
            output.append(row["id"])

        output.append(row["decimals"])
        output.append(int(row["is_abstract_currency"]))
        output.append(int(row["starknet_address"]))
        output.append(int(row["ethereum_address"]))
    return output


def _format_pairs(pairs: Dict[str, Union[int, str]]) -> List[str]:
    # TODO (rlkelly): use marshmallow to format
    output = []
    for row in pairs:
        for key in ["id", "quote_currency_id", "base_currency_id"]:
            if isinstance(row[key], str):
                output.append(str_to_felt(row[key]))
            else:
                output.append(row[key])
    return output


async def deploy_contract(cli_config, contract_name):
    import configparser

    from empiric.cli import net

    config_parser = configparser.ConfigParser()
    config_parser.read(cli_config)

    gateway_url, chain_id = config.validate_config(cli_config)
    client = net.init_client(gateway_url, chain_id)
    account_client = net.init_account_client(client, cli_config)

    compiled_contract_path = Path(
        config_parser["CONFIG"].get("contract-path", config.COMPILED_CONTRACT_PATH)
    )
    compiled_contract = (compiled_contract_path / f"{contract_name}.json").read_text(
        "utf-8"
    )

    declare_result = await Contract.declare(
        account_client,
        compiled_contract=compiled_contract,
    )
    await declare_result.wait_for_acceptance()

    deployment_result = await declare_result.deploy(
        constructor_args=[],
    )
    await deployment_result.wait_for_acceptance()

    contract_address = deployment_result.deployed_contract.address
    return contract_address



import os
import re
from argparse import ArgumentParser

from starkware.starknet.compiler.compile import compile_starknet_files


# Helper func
def extract_arg_name_and_type(string):
    if ":" in string:
        name, type = string.split(":")
        name = name.strip()
        type = type.strip()
    else:
        name = string.strip()
        type = "felt"  # Default type in Cairo

    return name, type


def check_interface(file_path, contract_filename, cairo_path=None):
    # Assumes convention that if contract is called "Contract.cairo"
    # then interface is "IContract.cairo"
    interface_filename = "I" + contract_filename

    # 1. Compile Contract.cairo to get the ABI
    compiled = compile_starknet_files(
        [os.path.join(file_path, contract_filename)], cairo_path=cairo_path.split(",")
    )
    functions = {d["name"]: d for d in compiled.abi if d["type"] == "function"}

    # 2. Check that the interface matches the ABI
    with open(os.path.join(file_path, interface_filename), "r") as f:
        contents = f.readlines()
        for i, line in enumerate(contents):
            # Check that namespace name is identical to interface filename
            if line == "@contract_interface":
                assert (
                    contents[i + 1]
                    == f"namespace {interface_filename.split('.')[0]} {{"
                ), "Wrong namespace name, doesn't match interface filename"

            # Check interface is implemented (same functions, same function signatures)
            if line.strip().startswith("func"):
                match = re.search(r"func .+?\(", line).group()
                function_name = match[5:-1]
                try:
                    abi = functions[function_name]
                except KeyError:
                    print(
                        f"Function {function_name} not found in the ABI, but is defined in the interface"
                    )

                # Gather complete function signature
                function_signature = ""
                j = 0
                while True:
                    if contents[i + j].strip() == "}":
                        break
                    function_signature += contents[i + j].strip()
                    j += 1

                # Check that function signature matches (inputs)
                inputs = re.search(r"\(.*?\)", function_signature).group()[1:-1]
                if not inputs:  # inputs is empty
                    assert (
                        abi["inputs"] == []
                    ), f"Function {function_name} has empty inputs in the interface, but not in the ABI"
                else:
                    inputs_list = inputs.split(",")
                    if inputs_list[-1] == "":
                        inputs_list = inputs_list[:-1]

                    assert len(inputs_list) == len(
                        abi["inputs"]
                    ), f"Function {function_name} has different lengths of inputs in the interface and implementation: {inputs} and {abi['inputs']}"
                    for arg, abi_input in zip(inputs_list, abi["inputs"]):
                        name, type = extract_arg_name_and_type(arg)
                        assert (
                            name == abi_input["name"]
                        ), f"Function {function_name} argument {name} and {abi_input['name']} do not match"
                        assert (
                            type == abi_input["type"]
                        ), f"Function {function_name} types {type} and {abi_input['type']} do not match"

                # Check that function signature matches (outputs)
                if re.search(r"-> \(.*?\)", line) is not None:  # There are outputs
                    outputs = re.search(r"-> \(.*?\)", line).group()[4:-1]

                    no_outputs = len(abi["outputs"]) == 0 and outputs == ""
                    matching_num_outputs = len(outputs.split(",")) == len(
                        abi["outputs"]
                    )

                    assert (
                        no_outputs or matching_num_outputs
                    ), f"Function {function_name} has different lengths of outputs in the interface and implementation: {outputs} and {abi['outputs']}"
                    for interface_output, abi_output in zip(
                        outputs.split(","), abi["outputs"]
                    ):
                        name, type = extract_arg_name_and_type(interface_output)

                        assert (
                            name == abi_output["name"]
                        ), f"Function {function_name} argument {name} and {abi_output['name']} do not match"
                        assert (
                            type == abi_output["type"]
                        ), f"Function {function_name} types {type} and {abi_output['type']} do not match"


if __name__ == "__main__":
    # Search for all files that are Contract.cairo
    # recursively within the "contracts" folder
    parser = ArgumentParser()
    parser.add_argument(
        "--cairo-path",
        default="contracts",
        help="Path to Cairo files, used to search for files to check and for relative imports when compiling.",
    )
    args = parser.parse_args()

    file_paths = []
    contract_filenames = []
    for path in args.cairo_path.split(","):
        for folder, _, files in os.walk(path):
            for file in files:
                if (
                    file.endswith(".cairo")
                    and not (file.startswith("I") and file[1].isupper())
                    and file[0].isupper()
                ):
                    file_paths.append(os.path.join(folder))
                    contract_filenames.append(file)

    for file_path, contract_filename in zip(file_paths, contract_filenames):
        print(f"Checking file {os.path.join(file_path, contract_filename)}")
        check_interface(file_path, contract_filename, args.cairo_path)



import time

import typer
from empiric.cli import config, net
from empiric.cli.utils import coro
from empiric.core import SpotEntry
from empiric.core.contract import wait_for_received
from empiric.core.utils import str_to_felt
from empiric.publisher import EmpiricPublisherClient
from empiric.publisher.assets import EMPIRIC_ALL_ASSETS
from empiric.publisher.fetchers import CexFetcher

app = typer.Typer(help="contract deployment utilities")


@app.command()
@coro
async def run(
    publisher: str = "empiric", delay: int = 30, config_path=config.DEFAULT_CONFIG
):
    """run a publisher locally"""
    client = net.init_empiric_client(config_path)
    EmpiricPublisherClient.convert_to_publisher(client)
    client.add_fetcher(CexFetcher(EMPIRIC_ALL_ASSETS, publisher))
    while True:
        start = time.time()
        _entries = await client.fetch()
        typer.echo(f"publishing {len(_entries)} entries")
        invocations = await client.publish_many(_entries, pagination=100)
        await invocations[0].wait_for_acceptance()

        typer.echo("setting checkpoint")
        res = await client.set_checkpoints(
            [int(entry.pair_id) for entry in _entries if isinstance(entry, SpotEntry)],
        )

        typer.echo(str(res.hash))
        typer.echo(str(await res.wait_for_acceptance()))
        now = time.time()
        if now - start < delay:
            time.sleep(delay - (now - start))
